#include "spiRam.h"

static volatile uint8_t spiMode;
static uint8_t spiRamAddr[3];
static uint8_t spiAddrIndex;
static uint16_t spiBytesRemaining;
static uint8_t *spiDramPtr;

void spiRam::start(void) {
  digitalWrite(SPI_SS, HIGH); // deselect RAM chip
  pinMode(SPI_SS, OUTPUT);
  DDRB |= 0x28; // port B bit 5 (D13 SCK) and bit 3 (D11 MOSI) are outputs
  DDRB &= 0xEF; // port B bit 4 (D12 MISO) is an input
  SPCR = (1 << SPE) | (1 << MSTR); // enable SPI in master mode
  SPSR = 0x01; // set x2 speed to give fastest possible SPI clock of fOsc/2 (8 MHz)
  
  digitalWrite(SPI_SS, LOW); // select RAM chip
  spiMode = 1; // start of configuration
  SPCR |= (1 << SPIE); // enable SPI interrupts
  SPDR = 0x01; // begin write to mode register to configure sequential mode
  while (spiMode) // wait for SPI to return to idle mode
    ;
}

void spiRam::writeSpi(uint8_t *src, uint32_t address, uint16_t count) { // src is pointer to normal Arduino SRAM
  startTransmission(src, address, count, 2, 3);
}

void spiRam::readSpi(uint8_t *dest, uint32_t address, uint16_t count) { // dest is pointer to normal Arduino SRAM
  startTransmission(dest, address, count, 3, 4);
}

void spiRam::startTransmission(uint8_t *ptr, uint32_t address, uint16_t count, uint8_t opcode, uint8_t mode) {
  spiRamAddr[0] = (uint8_t)((address >> 16) & 0x000000FFUL);
  spiRamAddr[1] = (uint8_t)((address >> 8) & 0x000000FFUL);
  spiRamAddr[2] = (uint8_t)(address & 0x000000FFUL);
  spiAddrIndex = 0;  // number of address bytes still to send
  spiBytesRemaining = count;
  spiDramPtr = ptr;
  digitalWrite(SPI_SS, LOW); // select RAM chip
  spiMode = mode;
  SPCR |= (1 << SPIE); // enable SPI interrupts
  SPDR = opcode; // begin operation
  // if you know that your code won't overwrite the transmit buffer before it's sent
  // or use the receive buffer before it's filled
  // or start another read/write operation before the current one is complete
  // then you don't need this busy wait loop.
  while (spiMode) // wait for SPI to return to idle mode
    ;  
}

ISR(SPI_STC_vect) { // SPI serial transfer interrupt
  switch (spiMode) {
    case 1: // first byte of configuring sequential mode sent
      SPDR = 0x40; // write 2nd (and last) byte of configuration message - setting mode to sequential
      spiMode = 2;
      break;
    case 2: // final byte of packet sent or received
      SPCR &= ~(1 << SPIE); // disable SPI interrupts
      digitalWrite(SPI_SS, HIGH); // deselect RAM chip
      spiMode = 0; // back to idle
      break;
    case 3: // sending address for write operation
    case 4: // sending address for read operation
      SPDR = spiRamAddr[spiAddrIndex++];
      if (spiAddrIndex == 3)
        spiMode += 2;
      break;
    case 5: // writing bytes to SPI ram
      SPDR = *spiDramPtr++;
      if (--spiBytesRemaining == 0)
        spiMode = 2;
      break;
    case 6: // discard first byte received after sending address
      SPDR ^= 0xFF; // read/write discards first byte and keeps SPI active
      spiMode = 7;
      break;
    case 7: // reading bytes from SPI ram
      SPDR = 0; // dummy data to keep SPI active for receive
      *spiDramPtr++ = SPDR;
      if (--spiBytesRemaining == 0)
        spiMode = 2;
      break;
  }
}
